use core::net::{Ipv4Addr, SocketAddrV4};

use log::debug;

use super::atem_packet::{AtemPacket, AtemPacketParseError, PacketFlag};

pub const ATEM_PORT: u16 = 9910;
pub const SYN_PAYLOAD: [u8; 8] = [0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00];

#[derive(Debug, PartialEq, Eq)]
pub enum AtemSocketState {
    Closed,
    StateDump,
    Established { session_id: u16 },
}

/// An abstraction over an async UDP socket
pub trait SocketBackend {
    type SendError;
    type RecvError;
    async fn send_to(&mut self, data: &[u8], addr: SocketAddrV4) -> Result<(), Self::SendError>;
    async fn recv(&mut self, data: &mut [u8]) -> Result<usize, Self::RecvError>;
}

#[cfg(feature = "std")]
pub mod tokio {
    use core::net::Ipv4Addr;

    use tokio::net::UdpSocket;

    use super::{AtemSocket, SocketBackend};

    #[derive(Debug)]
    pub struct TokioBackend {
        sock: UdpSocket,
    }

    impl TokioBackend {
        pub async fn new(addr: Ipv4Addr) -> Result<AtemSocket<Self>, std::io::Error> {
            let b = Self {
                sock: UdpSocket::bind("0.0.0.0:0").await?,
            };
            Ok(AtemSocket::new_with_backend(addr, b))
        }
    }

    impl SocketBackend for TokioBackend {
        type SendError = std::io::Error;
        type RecvError = std::io::Error;

        async fn send_to(
            &mut self,
            data: &[u8],
            addr: core::net::SocketAddrV4,
        ) -> Result<(), Self::SendError> {
            self.sock.send_to(data, addr).await?;
            Ok(())
        }

        async fn recv(&mut self, data: &mut [u8]) -> Result<usize, Self::RecvError> {
            let (cnt, _) = self.sock.recv_from(data).await?;
            Ok(cnt)
        }
    }
}

#[cfg(feature = "embassy-net")]
pub mod enet {
    use core::net::{Ipv4Addr, SocketAddrV4};

    use embassy_net::udp::{RecvError, SendError, UdpSocket};

    use super::{AtemSocket, SocketBackend};

    pub struct EnetBackend<'a> {
        sock: UdpSocket<'a>,
    }

    impl core::fmt::Debug for EnetBackend<'_> {
        fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
            write!(f, "EnetBackend")
        }
    }

    impl<'a> EnetBackend<'a> {
        pub fn new(addr: Ipv4Addr, sock: UdpSocket<'a>) -> AtemSocket<Self> {
            let b = Self { sock };
            AtemSocket::new_with_backend(addr, b)
        }
    }

    impl SocketBackend for EnetBackend<'_> {
        type SendError = SendError;
        type RecvError = RecvError;

        async fn send_to(
            &mut self,
            data: &[u8],
            addr: SocketAddrV4,
        ) -> Result<(), Self::SendError> {
            self.sock.send_to(data, addr).await
        }

        async fn recv(&mut self, data: &mut [u8]) -> Result<usize, Self::RecvError> {
            let (count, _) = self.sock.recv_from(data).await?;
            Ok(count)
        }
    }
}

#[derive(Debug)]
#[cfg_attr(feature = "defmt", derive(defmt::Format))]
pub enum AtemError<I: SocketBackend> {
    /// Tried to do something but the socket is closed
    SocketClosed,
    /// There was an error from the underlying socket while receiving
    Recv(I::RecvError),
    /// There was an error from the underlying socket while sending
    Send(I::SendError),
    /// We received an invalid ATEM protocol packet
    BadPacket(AtemPacketParseError),
    /// The switcher denied our connection attempt.
    SwitcherDeniedConnect,
    /// There was some nonspecific protocol error
    ProtocolError,
}

impl<I> From<AtemPacketParseError> for AtemError<I>
where
    I: SocketBackend,
{
    fn from(value: AtemPacketParseError) -> Self {
        Self::BadPacket(value)
    }
}

pub struct AtemSocket<I: SocketBackend> {
    inner: I,
    addr: SocketAddrV4,
    state: AtemSocketState,
    local_seq_num: u16,
}

impl<I> AtemSocket<I>
where
    I: SocketBackend,
{
    fn new_with_backend(addr: Ipv4Addr, backend: I) -> Self {
        Self {
            inner: backend,
            addr: SocketAddrV4::new(addr, ATEM_PORT),
            state: AtemSocketState::Closed,
            local_seq_num: 0,
        }
    }

    /// Connect to the ATEM, performing the connection handshake.
    pub async fn connect(&mut self) -> Result<(), AtemError<I>> {
        // Reset state
        self.state = AtemSocketState::Closed;
        self.local_seq_num = 0;
        let mut buf = [0u8; 20];
        // FIXME: this should be random
        let initial_sid = 0x53ab;

        // Step 1: Send the "SYN" with a random session id and magic payload

        let (mut p, _) = AtemPacket::init(
            &mut buf,
            PacketFlag::NewSessionId.into(),
            initial_sid,
            self.local_seq_num,
            Some(&SYN_PAYLOAD),
        )
        .unwrap();
        p.set_remote_seq_num(0x3a);

        self.inner
            .send_to(p.inner(), self.addr)
            .await
            .map_err(|e| AtemError::Send(e))?;

        debug!("Connect step 1: sent SYN");

        // Step 2: Wait for the response from the ATEM.
        // this should also have the SYN flag set and have the same session id that we sent.
        // The first byte of the payload should be 0x02 - if it's 0x04 we should restart
        // the connection

        self.inner
            .recv(&mut buf)
            .await
            .map_err(|e| AtemError::Recv(e))?;

        let ack = AtemPacket::new_checked(buf)?;
        debug!("RX: {}", ack);
        if ack.flags() != PacketFlag::NewSessionId {
            return Err(AtemError::ProtocolError);
        }
        if ack.session_id() != initial_sid {
            return Err(AtemError::ProtocolError);
        }
        match ack.body()[0] {
            0x02 => {
                // all good
                debug!("Connect step 2: got reply");
            }
            0x04 => {
                // Restart connection
                return Err(AtemError::SwitcherDeniedConnect);
            }
            _ => {
                return Err(AtemError::ProtocolError);
            }
        }

        // Step 3: Send an ACK for the packet we just received.
        self.ack(initial_sid, ack.local_sequence_number()).await?;
        debug!("Connect step 3: sent ACK");
        self.state = AtemSocketState::StateDump;

        Ok(())
    }

    async fn ack(&mut self, sid: u16, ack_num: u16) -> Result<(), AtemError<I>> {
        let mut buf = [0u8; 12];
        let (mut p, _) = AtemPacket::init(
            &mut buf,
            PacketFlag::AckReply.into(),
            sid,
            // N.B: local seq num is set to 0 in ACKs
            0,
            None,
        )
        .unwrap();
        p.set_ack_num(ack_num);
        debug!("TX: {}", p);
        self.inner
            .send_to(p.inner(), self.addr)
            .await
            .map_err(|e| AtemError::Send(e))?;
        Ok(())
    }

    /// Receive a packet from the ATEM. Automatically handles sending ACKs for received packets.
    pub async fn recv<'a>(
        &mut self,
        buf: &'a mut [u8],
    ) -> Result<AtemPacket<&'a [u8]>, AtemError<I>> {
        if self.state == AtemSocketState::Closed {
            return Err(AtemError::SocketClosed);
        }
        let size = self.inner.recv(buf).await.map_err(|e| AtemError::Recv(e))?;
        let p: AtemPacket<_> = buf[0..size].try_into()?;
        debug!("RX: {}", p);
        match self.state {
            AtemSocketState::StateDump => {
                // We've just connected and the state's being dumped to us.
                if p.body().len() == 0 {
                    // State dump is done.
                    self.state = AtemSocketState::Established {
                        session_id: p.session_id(),
                    };
                    debug!("ATEM state dump done, session ID is {}", p.session_id());
                    self.ack(p.session_id(), p.local_sequence_number()).await?;
                }
            }
            AtemSocketState::Established { session_id } => {
                if p.flags().contains(PacketFlag::AckRequest) {
                    self.ack(session_id, p.local_sequence_number()).await?;
                }
            }
            AtemSocketState::Closed => unreachable!(),
        }
        Ok(p)
    }

    /// Send a packet to the ATEM. The ATEM will respond with an ACK - it's up to the caller to handle packet retransmission
    pub async fn send<T: AsRef<[u8]>>(&mut self, p: AtemPacket<T>) -> Result<(), AtemError<I>> {
        self.inner
            .send_to(p.inner().as_ref(), self.addr)
            .await
            .map_err(|e| AtemError::Send(e))
    }
}
